; Random Pac-Man Style Maze Generator
; For ZX Spectrum
;
; by Jason J Railton
; June 2020

include "Macros.asm"

ENTRY_POINT		    EQU 32768

; Start with the MAZE GRID.
; Grid is 12x10 cells, to absorb over-run checks:  'X'.
; The maze outline is an 11x8 grid inside this:    '-+|'.
; The ghost-house and tunnels are placed first:    '='
; The inner maze walls are filled in randomly in a 5x6 grid on the left-hand side: '?',
;  then mirrored about the centre line: '.'
;
; X X X X X.X.X X X X X X
; + - - - -.-.- - - - + X 
; | ? ? ? ?.?.        | X 
; | ?      . .        | X 
; | ?     =.=.=       | X 
; | ?     =.=.=       | X 
; | ?      . .        | X 
; | ?      .?.        | X 
; + - - - -.-.- - - - + X 
; X X X X X.X.X X X X X X
;
; Grid data is a row of 2 x bytes, 0/1 for solid block, then 0/1 for wall to the next block to the right.
; Below that another row of 2x bytes, 0/1 for solid wall down to next block, and one dummy byte.
;
; MAZE_MAP is 32x22 grid of UDGs to draw the maze.
; Maze passageways appear two characters wide, but only the upper-left character is the passageway.
; The lower and right characters are treated as solid, as if the walls were two characters thick
;  instead of one. This makes negotiating the maze a simpler calculation, since we can treat it as
;  one-character wide passageways and just overdraw the sprites to 2x2 characters.
; The dots are actually in the bottom-right corner of the passageway UDGs rather than split across
;  2x2 characters, and all the sprites are actually just 14x14 in the top-left of their 16x16 size.
;

;----------------------------------------

CHAR_EMPTY_SPACE    EQU 0
CHAR_DOT            EQU 6
CHAR_PILL_A         EQU 3
CHAR_PILL_B         EQU 3
CHAR_SOLID_BLACK    EQU 1
CHAR_SOLID_PILL_A   EQU 2
CHAR_SOLID_PILL_B   EQU 2
CHAR_WALL_ZERO      EQU 7
CHAR_WALL_X         EQU 5
CHAR_WALL_Y         EQU 4

;----------------------------------------

; Random tuning:
		SEED EQU 8
		
		
;----------------------------------------

WORKSPACE_START     EQU 50000

; - - - - - - - -

MAZE_GRID           EQU 50000
MAZE_GRID_SIZE      EQU 480     ; 12 x 10 x 4

MAZE_MAP            EQU MAZE_GRID+MAZE_GRID_SIZE
MAZE_MAP_SIZE       EQU 704     ; 32 x 22 x 1


WORKKSPACE_END      EQU MAZE_MAP+MAZE_MAP_SIZE

; - - - - - - - -

WORKSPACE_SIZE      EQU WORKKSPACE_END-WORKSPACE_START


;----------------------------------------

org ENTRY_POINT

; - - - - - - - -
    ld sp,65536

    ld a,24    
    call cls
    ld a,3
    out (254),a
    call paint_q    
    
restart_hold
    ld bc,(random_seed)
wait_init_unkey
    inc bc
    in a,(254)
    and 31
    xor 31
    jp nz,wait_init_unkey
    
wait_init_key
    inc bc
    inc a
    in a,(254)
    and 31
    xor 31
    jp z,wait_init_key
    ld (random_seed),bc
    
    ld a,7
    call cls
    xor a
    out (254),a
    
    call plot_logo

    call clear_workspace

    call plot_perimeter
    
    ld hl,reseq_ghost_house
    call rewrite_sequence    
    
    ld hl,reseq_top_table
    call rewrite_random    
    
    ld hl,reseq_bottom_table
    call rewrite_random    
    
    ld hl,reseq_tunnel_table
    call rewrite_random    
    
    call fit_shapes
    
    ; Shouldn't be necessary, but just in case:
    call finalise_maze
    
    call mirror_grid
    
    call grid_to_map
    
    call map_to_screen    
    
    
    jp restart_hold
    
; - - - - - - - -

wait_forever
    jp wait_forever

    
;================================================================
; Maze Grid Manipulation Functions
;================================================================
    
plot_perimeter

    ; Top and bottom horizontal walls:
        
    ld hl,MAZE_GRID + (1*48)
    ld de,MAZE_GRID + (8*48)
    ld a,1
    
    ld b,10
plot_perimeter_x_loop
    ld (hl),a               ; Solid block
    inc hl
    ld (hl),a               ; Wall off to right
    inc hl
    
    ld (de),a               ; Solid block
    inc de    
    ld (de),a               ; Wall off to right
    inc de   
     
    djnz plot_perimeter_x_loop

    ld (hl),a               ; Last solid block
    ld (de),a               ; Last solid block

        
    ; Left vertical wall (gets mirrored):
            
    ld hl,MAZE_GRID + (1*48)
    ld bc,24
    
    ld a,14                 ; Seven rows * 2
plot_perimeter_y_loop

    ld (hl),1               ; Solid block one row, wall downward on next row
    add hl,bc
    
    dec a
    jp nz,plot_perimeter_y_loop
        
    ret

    
;----------------------------------------

mirror_grid

    ld hl,MAZE_GRID + (1*48)
    ld de,MAZE_GRID + (1*48) + 20
    
    ld c,15
    mirror_grid_loop_y

    ld b,10
    mirror_grid_loop_x
    
    ld a,(hl)
    ld (de),a
    inc hl
    dec de
    
    djnz mirror_grid_loop_x
    
    ld a,c
        ld bc,14        ; +24 -10 already covered
        add hl,bc
        
        ex de,hl
        ld bc,34        ; +24 +10 already decremented
        add hl,bc
        ex de,hl
    ld c,a
    dec c
    jp nz,mirror_grid_loop_y
    
    ret
    
    
;----------------------------------------

; Rewrite MAZE GRID with a choice of sequences from a table:

rewrite_random

    call random

    and (hl)        ; First byte of table is AND mask for random number
    inc hl
    
    ld b,0
    ld c,a
    sla c
    rl b
    add hl,bc       ; Point to table entry
    
    ld a,(hl)
    inc hl
    ld h,(hl)
    ld l,a          ; Get table entry

;   call rewrite_sequence     ; Action changes...
;   ret

; - - - - - - - -

; Rewrite MAZE GRID with a pre-defined sequence:

rewrite_sequence
    ld a,(hl)
    inc hl
    push af     ; Data byte. Return with this in A

    ld b,(hl)   ; Number of actions
    inc hl
rewrite_seq_loop_1
    push bc
    
    ; Get start position of action:
    
    ld a,(hl)   ; A = X
    inc hl
    ld b,(hl)   ; B = Y
    inc hl
    
    call maze_grid_address
    
    ; Action a row of changes:
    
    ld b,(hl)   ; Number of cells to re-write for this action
    inc hl
rewrite_seq_loop_2
    
    ld a,(hl)   ; Re-write central block of cell
    inc hl
    ld (iy),a

    ld a,(hl)   ; Re-write wall off to right
    inc hl
    ld (iy+1),a
    
    ld a,(hl)   ; Re-write wall downward
    inc hl
    ld (iy+24),a
    
    inc iy
    inc iy      ; Next cell over
    
    djnz rewrite_seq_loop_2    ; Complete action
    
    
    pop bc
    djnz rewrite_seq_loop_1    ; Next action
    
    
    pop af      ; Recover data byte in A before returning
    ret


;----------------------------------------
; Calculate maze grid address in IY.
; On entry: A=X, B=Y
; Destroys A, B, C

maze_grid_address    
    ;A=X, B=Y    
    ld c,0      ; BC = Y *256
    
    ld iy,MAZE_GRID
    sra b
    rr c        ; BC = Y * 128
    sra b
    rr c        ; BC = Y * 64
    sra b
    rr c        ; BC = Y * 32
    add iy,bc   ; IY = MAZE_GRID + Y*32
    sra b
    rr c        ; BC = Y * 16
    add iy,bc   ; IY = MAZE_GRID + Y*48
    
    rla
    ld b,0
    ld c,a
    add iy,bc   ; IY = MAZE_GRID + Y*48 + X*2

    ret
    
    
;================================================================
; Maze Shape-Fitting Functions
;================================================================

fit_shapes
    ld b,1          ; Maze column counter (from 1 to 4)  
    
fit_shapes_column_loop    

    ld c,1          ; Row counter (from 1 to 6)
    
fit_shapes_row_loop    

    push bc
    
    ; Calculate the maze address in IY:
    ld a,b          ; X-position
    
    ld b,c          ; Y-position.
    
    call maze_grid_address       ;A=X, B=Y, get address in IY
    
    
    ; Clear all the 'tried this shape' flags:
    
    ld hl,shape_list+1
    ld de,ONE_SHAPE_SIZE
    ld b,SHAPE_COUNT
fit_shapes_reset_loop
    ld (hl),0
    add hl,de
    djnz fit_shapes_reset_loop
 
       
   ; Pick a shape at random, and try to fit it.
   ; If it won't fit, try the next shape in the list and so-on:
    
    call random         ; Pick a shape
    and $F0
    
    
    ; Get address of shape data:
    
    ld b,0
    ld c,a              ; BC = 16 x shape_ID
    ld hl,shape_list
    add hl,bc
    add hl,bc           ; HL = shape_list + 32 x shape_ID
    
    ld b,(hl)           ; B = initial choice of shape_ID from list

    
    ; Repeat to pick a shape that hasn't already been tried:
    
    fit_shapes_not_used_loop    
        inc hl    
        ld a,(hl)               ; Already-tried flag of current shape
        
        and a                   ; Has this shape already been tried here?
        jp z,fit_shapes_try     ; If not, try to fit it
        
    fit_shapes_failed_re_entry
    
        inc hl                  ; Get address of next shape to try...
        ld e,(hl)
        inc hl
        ld d,(hl)
        ex de,hl                ; HL = next shape
        
        ld a,(hl)               ; Get it's ID
        cp b                    ; Have we gone round the list and back to where we started?
        JPEQ fit_shapes_next    ; If so, abort trying to fit a shape here
        
    jp fit_shapes_not_used_loop ; Otherwise, try to fit that next shape
    
    
; - - - - - - - -
    
fit_shapes_try
    ld (hl),1               ; Flag this shape as trying to fit

    call place_shape_xor    ; Place shape and check for collisions
    and a                   ; Were there collisions with existing maze?
    jp nz,fit_shapes_undo   ; If so, remove shape and try again
        
    call survey_maze        ; Check maze for isolated, unconnnectable cells
    and a                   ; Were there any?
    jp nz,fit_shapes_undo   ; If so, remove shape and try again
    
    
    ; If shape fitted OK, move onto the next cell in
    ;  the maze and try to fit a shape:

fit_shapes_next
    pop bc
    inc c
    ld a,c
    cp 7
    JPLT fit_shapes_row_loop
    
    inc b
    ld a,b
    cp 5
    JPLT fit_shapes_column_loop
    
    ret
    
    
; - - - - - - - -    
; Remove a placed shape (with XOR) and go back
; to try another shape:    

fit_shapes_undo 
    call place_shape_xor
    
    jp  fit_shapes_failed_re_entry   
    
    
;----------------------------------------
; Fit the shape (HL) into the maze (IY)
; And return any collisions in A:

place_shape_xor
    push bc
    push de
    push hl
    push iy
    xor a
    
    inc hl
    inc hl
    inc hl                  ; Start of wall data
    
    ld c,3              ;X-count
place_shape_xor_loop_x   
    ld b,3              ;Y-count
place_shape_xor_loop_y
    push bc
    ld e,a      ; Cache collisions in E
    
    ld a,(iy+0)
    ld d,a      ; Get existing byte in D
    and (hl)    ; Calculate any collision in A
    or e
    ld e,a      ; OR into E to collate any collisions
    ld a,d
    xor (hl)    ; XOR wall data and replace in maze
    ld (iy+0),a
    inc hl
    
    ld a,(iy+1)
    ld d,a
    and (hl)    ; Calculate any collision in A
    or e
    ld e,a
    ld a,d
    xor (hl)
    ld (iy+1),a
    inc hl
    
    ld a,(iy+24)
    ld d,a
    and (hl)    ; Calculate any collision in A
    or e
    ld e,a
    ld a,d
    xor (hl)
    ld (iy+24),a
    inc hl
    
    inc iy
    inc iy
    
    pop bc
    ld a,e      ; Collision record back into A
    djnz place_shape_xor_loop_y
    
    ld de,48-6
    add iy,de
    
    dec c
    jp nz,place_shape_xor_loop_x
    
    pop iy
    pop hl
    pop de
    pop bc
    ret 
     
    
;----------------------------------------
; Survey the maze for any wall cells that are
; empty, but have walls on all four sides, and
; so can't be connected up to make a shape.

survey_maze
    push iy
    push de
    push bc
    xor a
    
    ld iy,MAZE_GRID+48+48+2
    ld de,48-8                  ; Next row step
    
    ld c,6                      ; Row count 
survey_maze_loop_x
    ld b,4                      ; Column count
survey_maze_loop_y
    ld a,(iy)
    and a                       ; Is this maze cell a set wall?
    jp nz,survey_maze_next      ; If so, that's fine
    
    ld a,1                      ; If not, check surrounding cells for all solid
    and (iy-2)
    and (iy+2)
    and (iy-48)
    and (iy+48)
    jp nz,survey_maze_return    ; And if all fenced in, return with a 1 to signal a problem
    
survey_maze_next
    inc iy
    inc iy
    djnz survey_maze_loop_y
    
    add iy,de
    dec c
    jp nz,survey_maze_loop_x
    
    xor a                       ; Return with a 0, to signal no problem

survey_maze_return
    pop bc
    pop de
    pop iy
    ret

    
;----------------------------------------
; Set all wall cells in a maze, even if not joined
; to anything. Shouldn't need this step, but just
; in case. This will leave tell-tale single wall
; spots in the maze.

finalise_maze
    ld hl,MAZE_GRID+48+48+2
    ld de,48-8                  ; Next row step
    ld c,6                      ; Row count 
finalise_maze_loop_x
    ld b,4                      ; Column count
finalise_maze_loop_y
    ld (hl),1
    inc hl
    inc hl
    djnz finalise_maze_loop_y
    add hl,de
    dec c
    jp nz,finalise_maze_loop_x
    
    ret


;================================================================
; Maze Translation Functions
;================================================================
    
grid_to_map
    ld ix,MAZE_GRID+48
    ld iy,MAZE_MAP
    
    ld c,8
grid_to_map_loop_y
    
    ld b,11
grid_to_map_loop_x

    ; Check block:

    ld a,(ix)
    and a
    jp z,grid_to_map_skip_1    
        ld a,CHAR_WALL_ZERO
        ld (iy),a
    grid_to_map_skip_1
    
    ; Check for right wall:

    ld a,(ix+1)
    and a
    jp z,grid_to_map_skip_2    
        ld a,CHAR_WALL_X
        ld (iy+1),a
        ld (iy+2),a
    grid_to_map_skip_2
    
    ; Check for down wall:

    ld a,(ix+24)
    and a
    jp z,grid_to_map_skip_3    
        ld a,CHAR_WALL_Y
        ld (iy+32),a
        ld (iy+64),a
    grid_to_map_skip_3
    

    inc ix
    inc ix      ; Next grid cell
    
    inc iy
    inc iy
    inc iy      ; Next map equivalent
    
    djnz grid_to_map_loop_x
    
    ld a,c
        
        ld bc,26
        add ix,bc   ; Next grid row = +48 -22 just done

        ld bc,63      ; Next screen equivalent = +96 -33 just done
        add iy,bc    
    
    ld c,a
    
    dec c
    jp nz,grid_to_map_loop_y

    ret


;----------------------------------------

map_to_screen
    ld hl,MAZE_MAP
    ld de,ATTRS

    ld c,22  
map_to_screen_loop_y

    ld b,31
map_to_screen_loop_x

    ld a,(hl)
    inc hl
    
    and 7
    rlca
    rlca
    rlca
    or 64
    ld (de),a
    inc e
    
    djnz map_to_screen_loop_x
    
    inc hl
    inc de
    
    dec c
    jp nz,map_to_screen_loop_y
    
    ret


;================================================================
; Reset Functions
;================================================================

clear_workspace
    ld hl,WORKSPACE_START
    ld de,WORKSPACE_START+1
    ld bc,WORKSPACE_SIZE-1
    ld (hl),0
    ldir
    
    ld hl,MAZE_MAP
    ld de,MAZE_MAP+1
    ld bc,MAZE_MAP_SIZE-1
    ld (hl),CHAR_SOLID_BLACK
    ldir
    
    ret
    
    
;----------------------------------------
	
cls
	ld hl,SCREEN
	ld de,SCREEN+1
	ld bc,SCREEN_SIZE-1
	ld (hl),0
	ldir
	
	inc hl
	inc de
	ld (hl),a
	ld bc,ATTRS_SIZE-1
	ldir
	
	ret
    
	
;================================================================
; Random Number
;================================================================

; Return random number in A, all other registers preserved:

random
	push de
	push hl
random_init
random_seed EQU random_init+1
	ld	de,SEED		;gets over-written with each call
	ld	a,d
	ld	h,e
	ld	l,253
	or	a
	sbc	hl,de
	sbc	a,0
	sbc	hl,de
	ld	d,0
	sbc	a,d
	ld	e,a
	sbc	hl,de
	jr	nc,rand_skip
	inc	hl
rand_skip
	ld	(random_seed),hl
	ld a,h
	xor l
	pop hl
	pop de
	ret


;================================================================
; Pre-Defined Re-Write Sequences
;================================================================
;
; DATA: info_byte, num_changes
;       [repeat num_changes:] X, Y, N, [repeat N:] Solid_Block, Horz_Wall, Vert_Wall 
;
; TABLE: random_number_AND_byte
;        [repeat AND result:] sequence_address


reseq_ghost_house
    db 0,  2
    db 4,4,3,  1,1,1,  1,1,0,  1,0,1    ; top row
    db 4,5,3,  1,1,0,  1,1,0,  1,0,0    ; bottom row

    
;----------------------------------------

reseq_top_table
    db 3        ; Four options, first is duplicated at end
    dw reseq_top_0, reseq_top_1, reseq_top_2, reseq_top_0
    
    reseq_top_0
        db 0,  3            ; Vertical bar down from top-centre
        db 5,1,1,  1,1,1
        db 5,2,1,  1,0,1
        db 5,3,1,  1,0,0
        
    reseq_top_1
        db 0,  2            ; T-bar down from top-centre. Only need half as mirrored
        db 4,2,2,  1,1,0,  1,1,1
        db 5,3,1,  1,0,0
        
    reseq_top_2
        db 0,  3            ; T-bar down from top-centre, plus side spike(s)
        db 3,1,1,  1,1,1
        db 3,2,3,  1,0,0,  1,1,0,  1,1,1
        db 5,3,1,  1,0,0

        
;----------------------------------------
        
reseq_bottom_table
    db 7        ; Eight options 
    dw reseq_bottom_0, reseq_bottom_1, reseq_bottom_2, reseq_bottom_3
    dw reseq_bottom_4, reseq_bottom_5, reseq_bottom_6, reseq_bottom_7
    
    ; Upper player start position (under Ghost House):
    
    reseq_bottom_0
        db 13,  2           ; Horizontal bar under Ghost House, spike below
        db 4,6,2,  1,1,0,  1,1,0
        db 5,7,1,  1,0,1
                
    reseq_bottom_1
        db 13,  2           ; T-bar under Ghost House
        db 4,6,2,  1,1,0,  1,1,1
        db 5,7,1,  1,0,0
                
    reseq_bottom_2
        db 13,  2           ; T-bar under Ghost House, two spikes
        db 4,6,2,  1,1,0,  1,1,1
        db 4,7,2,  1,0,1,  1,0,0
                
    reseq_bottom_3
        db 13,  2           ; Downward-C under Ghost House, spike below
        db 4,6,2,  1,1,1,  1,1,0
        db 4,7,2,  1,0,0,  1,0,1
                
    ; Lower player start position (One extra row under Ghost House):
    
    reseq_bottom_4
        db 16,  2           ; Horizontal bar under Ghost House, spike below
        db 4,6,2,  1,1,0,  1,1,0
        db 5,7,1,  1,0,1
                
    reseq_bottom_5
        db 16,  2           ; Downward-C under Ghost House, spike below
        db 4,6,2,  1,1,1,  1,1,0
        db 4,7,2,  1,0,0,  1,0,1
                
    reseq_bottom_6
        db 16,  2           ; Two horizontal bars under Ghost House
        db 4,6,2,  1,1,0,  1,1,0
        db 4,7,2,  1,1,0,  1,1,0
                
    reseq_bottom_7
        db 16,  2           ; Horizontal bars under Ghost House, two spikes
        db 4,6,2,  1,1,0,  1,1,0
        db 3,7,3,  1,0,1,  1,1,0,  1,1,0
              
                          
;----------------------------------------

reseq_tunnel_table
    db 15        ; Sixteen options, some repeated
    dw reseq_tunnel_0,  reseq_tunnel_1,  reseq_tunnel_2,  reseq_tunnel_3
    dw reseq_tunnel_4,  reseq_tunnel_5,  reseq_tunnel_6,  reseq_tunnel_7
    dw reseq_tunnel_8,  reseq_tunnel_9,  reseq_tunnel_10, reseq_tunnel_11
    dw reseq_tunnel_12, reseq_tunnel_13, reseq_tunnel_0,  reseq_tunnel_12
    
    reseq_tunnel_0
        db 0,   1           ; No tunnel, middle upper spike
        db 0,4,2,  1,1,1,  1,0,0
        
    reseq_tunnel_1
        db 0,   1           ; No tunnel, middle lower spike
        db 0,5,2,  1,1,1,  1,0,0
        
    reseq_tunnel_2
        db 0,   2           ; No tunnel, top / bottom spikes
        db 0,3,2,  1,1,1,  1,0,0
        db 0,6,2,  1,1,1,  1,0,0
        
    reseq_tunnel_3
        db 1,   3           ; Top tunnel, upper spike
        db 0,1,1,  1,1,0
        db 0,2,2,  1,1,1,  1,0,0
        db 0,4,2,  1,1,1,  1,0,0
        
    reseq_tunnel_4
        db 1,   3           ; Top tunnel, middle spike
        db 0,1,1,  1,1,0
        db 0,2,2,  1,1,1,  1,0,0
        db 0,5,2,  1,1,1,  1,0,0
        
    reseq_tunnel_5
        db 1,   3           ; Top tunnel, lower spike
        db 0,1,1,  1,1,0
        db 0,2,2,  1,1,1,  1,0,0
        db 0,6,2,  1,1,1,  1,0,0
        
    reseq_tunnel_6
        db 19,  2           ; Bottom tunnel, upper spike
        db 0,7,2,  1,1,0,  1,0,0
        db 0,3,2,  1,1,1,  1,0,0
        
    reseq_tunnel_7
        db 19,  2           ; Bottom tunnel, middle spike
        db 0,7,2,  1,1,0,  1,0,0
        db 0,4,2,  1,1,1,  1,0,0
        
    reseq_tunnel_8
        db 19,  2           ; Bottom tunnel, lower spike
        db 0,7,2,  1,1,0,  1,0,0
        db 0,5,2,  1,1,1,  1,0,0
        
    reseq_tunnel_9
        db 7,   2           ; Middle tunnel
        db 0,4,2,  1,1,0,  1,0,0
        db 0,5,2,  1,1,1,  1,0,0
        
    reseq_tunnel_10
        db 4,   2           ; Mid-upper tunnel
        db 0,3,2,  1,1,0,  1,0,0
        db 0,4,2,  1,1,1,  1,0,0
        
    reseq_tunnel_11
        db 10,  2           ; Mid-lower tunnel
        db 0,5,2,  1,1,0,  1,0,0
        db 0,6,2,  1,1,1,  1,0,0
        
    reseq_tunnel_12
        db 4,  3            ; Mid-upper tunnel, lower spike
        db 0,3,2,  1,1,0,  1,0,0
        db 0,4,2,  1,1,1,  1,0,0
        db 0,6,2,  1,1,1,  1,0,0
        
    reseq_tunnel_13
        db 10,  3           ; Mid-lower tunnel, upper  spike
        db 0,5,2,  1,1,0,  1,0,0
        db 0,6,2,  1,1,1,  1,0,0
        db 0,3,2,  1,1,1,  1,0,0
        
        
;================================================================
; Maze Component Shapes
;================================================================

; Data Format:
;    1b: Shape ID (0..15)
;    1b: Tried / not tried flag (modifiable)
;    2b: Address of next shape
;   27b: 9 x [ cell, right-wall, downward-wall ]
;    1b: Padding to 32b

; Note that all shapes have their top-left cell empty and unconnected.
; Shape fitting is really taking place centred on the cell half-way-down
; the left-hand-side. So there is no point having a wall set in the top-left
; cell as that will overlap with the previous attempt to fit a shape.
; A wall set in the bottom-left is OK, because we won't have tried to fit a
; shape there yet (shape fitting is attempted in the maze from top to bottom).
 
ONE_SHAPE_SIZE      EQU 32
SHAPE_COUNT         EQU 17

shape_list
    shape_0     ;Cross +
        db 0, 0
        dw shape_1
        db 0,0,0,  1,0,1,  0,0,0
        db 1,1,0,  1,1,1,  1,0,0
        db 0,0,0,  1,0,0,  0,0,0
        db 0 

    shape_1     ;T-Up -'-
        db 1, 0
        dw shape_2
        db 0,0,0,  1,0,1,  0,0,0
        db 1,1,0,  1,1,0,  1,0,0
        db 0,0,0,  0,0,0,  0,0,0
        db 0 

    shape_2     ;T-Down -,-
        db 2, 0
        dw shape_3
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,0,  1,1,1,  1,0,0
        db 0,0,0,  1,0,0,  0,0,0
        db 0 

    shape_3     ;T-Left -|
        db 3, 0
        dw shape_4
        db 0,0,0,  1,0,1,  0,0,0
        db 1,1,0,  1,0,1,  0,0,0
        db 0,0,0,  1,0,0,  0,0,0
        db 0 

    shape_4     ;T-Right |-
        db 4, 0
        dw shape_5
        db 1,0,1,  0,0,0,  0,0,0
        db 1,1,1,  1,0,0,  0,0,0
        db 1,0,0,  0,0,0,  0,0,0
        db 0 

    shape_5     ;Wide-L, Down-Left ,--
        db 5, 0
        dw shape_6
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,1,  1,1,0,  1,0,0
        db 1,0,0,  0,0,0,  0,0,0
        db 0 

    shape_6     ;Wide-L, Up-Left '--
        db 6, 0
        dw shape_7
        db 0,0,0,  0,0,0,  0,0,0
        db 1,0,1,  0,0,0,  0,0,0
        db 1,1,0,  1,1,0,  1,0,0
        db 0 

    shape_7     ;S-Step Up _,-
        db 7, 0
        dw shape_8
        db 0,0,0,  1,1,1,  1,0,0
        db 1,1,0,  1,0,0,  0,0,0
        db 0,0,0,  0,0,0,  0,0,0
        db 0 

    shape_8     ;S-Step Down -,_
        db 8, 0
        dw shape_9
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,0,  1,0,1,  0,0,0
        db 0,0,0,  1,1,0,  1,0,0
        db 0        
        
    shape_9     ;3-Wide Bar ---
        db 9, 0
        dw shape_10
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,0,  1,1,0,  1,0,0
        db 0,0,0,  0,0,0,  0,0,0
        db 0 

    shape_10     ;Short-L Up-Right -'
        db 10, 0
        dw shape_11
        db 0,0,0,  1,0,1,  0,0,0
        db 1,1,0,  1,0,0,  0,0,0
        db 0,0,0,  0,0,0,  0,0,0
        db 0 

    shape_11     ;Short-L Down-Right -,
        db 11, 0
        dw shape_12
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,0,  1,0,1,  0,0,0
        db 0,0,0,  1,0,0,  0,0,0
        db 0 

    shape_12     ;Short-L Down-Left ,-
        db 12, 0
        dw shape_13
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,1,  1,0,0,  0,0,0
        db 1,0,0,  0,0,0,  0,0,0
        db 0 

    shape_13     ;Short-L Up-Left '-
        db 13, 0
        dw shape_14
        db 0,0,0,  0,0,0,  0,0,0
        db 1,0,1,  0,0,0,  0,0,0
        db 1,1,0,  1,0,0,  0,0,0
        db 0 

    shape_14     ;Horizontal Dash --
        db 14, 0
        dw shape_15
        db 0,0,0,  0,0,0,  0,0,0
        db 1,1,0,  1,0,0,  0,0,0
        db 0,0,0,  0,0,0,  0,0,0
        db 0 

    shape_15     ;Vertical Dash :
        db 15, 0
        dw shape_16
        db 0,0,0,  0,0,0,  0,0,0
        db 1,0,1,  0,0,0,  0,0,0
        db 1,0,0,  0,0,0,  0,0,0
        db 0 

    ; Further shapes can't be selected randomly, but may
    ; be needed to fill a particular gap:
        
    shape_16     ;3-Tall Bar |
        db 16, 0
        dw shape_0
        db 1,0,1,  0,0,0,  0,0,0
        db 1,0,1,  0,0,0,  0,0,0
        db 1,0,0,  0,0,0,  0,0,0
        db 0 

;================================================================

paint_q
    ld ix,ATTRS+32*11+15
    ld a,11001110b
    ld (ix-65),a
    ld (ix-64),a
    ld (ix-63),a
    ld (ix-34),a
    ld (ix-30),a
    ld (ix+01),a
    ld (ix+32),a
    ld (ix+96),a

    ret
    
;================================================================

plot_logo
    ld hl,logo_bin
    ld de,SCREEN+2048+2048+32*7
    ld a,8
    
    plot_logo_loop
        ldi
        ldi
        ldi
        ldi
        ldi
        ldi
        ldi
        inc d
        ld e,32*7
        dec a
    jp nz,plot_logo_loop
        
    ld de,ATTRS+32*23    
    ldi
    ldi
    ldi
    ldi
    ldi
    ldi
    ldi

    ret
    
        
logo_bin
    db 01111100b, 00000000b, 00000000b, 01111100b, 01100000b, 00010010b, 00000000b
    db 00010000b, 00000000b, 00000000b, 00010000b, 01010000b, 00010010b, 00000000b
    db 00010011b, 01110111b, 01100000b, 00010000b, 01010011b, 01010111b, 01110110b
    db 00010101b, 01000101b, 01010000b, 00010000b, 01100101b, 00010010b, 01010101b
    db 00010101b, 01110101b, 01010000b, 00010000b, 01010101b, 01010010b, 01010101b
    db 01010101b, 00010101b, 01010000b, 01010000b, 01010101b, 01010010b, 01010101b
    db 01110111b, 01110111b, 01010010b, 01110010b, 01010111b, 01010010b, 01110101b
    db 0,0,0,0,0,0,0    
    db 00+6, 64+6, 64+7, 00+7, 64+7, 64+6, 00+6
    
;================================================================

    db "Random Pac-Man Style Maze Generator",13
    db "For ZX Spectrum",13
    db 13
    db "by Jason J Railton",13
    db "June 2020",0

        
;================================================================
; Compilation End
;================================================================

end ENTRY_POINT

;================================================================
